package tracer;

import java.io.IOException;
import java.util.Random;

/**
 * Holds shading parameters.
 */
public class Material {
	
	public Material() {
		color = new Vec3( 1, 1, 1 );
		perlin = new Vec3( 1, 1, 1 );
		ambient = 0.2f;
		diffuse	= 0.8f;
		specular = 0.0f;
		specularPower = 16.0f;
		reflectance = 0.0f;
	}
	
	public Vec3 color;	
	public Vec3 perlin;
	public float ambient;
	public float diffuse;
	public float specular;
	public float specularPower;
	public float reflectance;
	public Vec3[] perlinArray;
	public int[] p;
	public Sampler2D sampler;
	
	public void parse( Parser p ) throws IOException {
		p.parseKeyword( "{" );
		while( !p.tryKeyword("}") && !p.endOfFile() ) {
			
			if( p.tryKeyword("color") ) {
				color.parse( p );
				perlin = color;
			} else if( p.tryKeyword("perlin") ) {
				color.parse( p );
				perlin.parse( p );
				setPerlinArrays();
			} else if( p.tryKeyword("ambient") ) {
				ambient = p.parseFloat();
			} else if( p.tryKeyword("diffuse") ) {
				diffuse = p.parseFloat();
			} else if( p.tryKeyword("specular") ) {
				specular = p.parseFloat();
			} else if( p.tryKeyword("specularpower") ) {
				specularPower = p.parseFloat();
			} else if( p.tryKeyword("reflectance") ) {
				reflectance = p.parseFloat();
			} else if( p.tryKeyword("texture") ) {
				{
					if( p.tryKeyword("debug") ){
						sampler = new DebugSampler();
					} else if( p.tryKeyword("nearest") ) {
						String filename = p.parseString();
						sampler = new NearestSampler(filename);
					} else if( p.tryKeyword("bilinear") ) {
						String filename = p.parseString();	
						sampler = new BilinearSampler(filename);
					}
				}
						
			} else {
				System.out.println( p.tokenWasUnexpected() );	
			}
			
		}
	}
	
	private void setPerlinArrays()
	{
		//perlinArray wordt gebruikt voor de random vectoren in op te slaan bij perlin noise.
		//Wordt pas met goede waarden gevuld als er daadwerkelijk een perlin texture is..
		perlinArray = new Vec3[256];
		for(int i=0; i<perlinArray.length; i++)
		{
			perlinArray[i] = Vec3.perlinVec();
		}
		
		//		p is de hasharray voor perlinfunctie. Dit is voor wat extra randomness.
		p = new int[perlinArray.length];
		Random randgen = new Random();
		
		//array vullen met waarden, zodat we later kunnen zien of er al een nieuwe waarde is
		//ingevoegd. aangezien de waarden in deze array de waarden 0 t/m n-1 hebben, waar n de lengte
		//van de perlinarray is, kunnen die nooit -1 worden en is dit correct
		for(int i=0; i<perlinArray.length; i++)
			p[i] = -1;
		
		//nu de permutatie in de array zetten, mbv van linear probing
		for(int i=0; i<perlinArray.length; i++)
		{
			int positie = randgen.nextInt(perlinArray.length);
			while(p[positie] != -1)
				positie = (positie+1)%perlinArray.length;
			p[positie] = i;
		}
	}
}