package tracer;

import java.io.IOException;

/**
 * 3D flat, onbounded, traceable thingy, represented by normal vector and distance from the origin
 */
public class Plane extends Traceable {

	Vec3 normal;
	float offset;
	Vec3 uDirection, vDirection;
	
	public Plane( Vec3 n, float o ) {
		normal = new Vec3(n);
		normal.normalize();
		offset = o;
		material = new Material();
		uDirection = new Vec3();
		vDirection = new Vec3();
	}
	public Plane() {
		normal = new Vec3(0,1,0);
		offset = 0.0f;
		material = new Material();
		uDirection = new Vec3();
		vDirection = new Vec3();
	}
	
	public void parse( Parser p ) throws IOException {
		p.parseKeyword( "{" );
		while( !p.tryKeyword("}") && !p.endOfFile() ) {
			
			if( p.tryKeyword("normal") ) {
				normal.parse( p );
			} else if( p.tryKeyword("offset") ) {
				offset = p.parseFloat();
			} else if( p.tryKeyword("material") ) {
				material.parse( p );
			} else if( p.tryKeyword("udirection") ) {
				uDirection.parse( p );
			} else if( p.tryKeyword("vdirection") ) {
				vDirection.parse( p );
			} else {
				System.out.println( p.tokenWasUnexpected() );	
			}
			
		}
		
		if (uDirection.isNewVec() && vDirection.isNewVec())
		{
			//Als U- en V-direction ontbreken, moeten we deze bepalen. We bepalen eerst de U-direction
			//en wegens efficientie gebruiker we voor de tijdlijke vector vdirection, die we na 
			//het zetten van udirection goed zetten. de tijdelijke v-direction moet ongelijk de normal zijn.
			vDirection = new Vec3(1 - normal.x, normal.y + 1, normal.z - 3);
			
			uDirection = normal.cross(vDirection);
			vDirection = normal.cross(uDirection);
			
			//nu voeren we een schaling uit, die optioneel is, zodat de textures niet erg klein worden
			uDirection.normalize();
			vDirection.normalize();
			uDirection = uDirection.times(.1f);
			vDirection = vDirection.times(.1f);
		}
		else
		{
			//als een van de 2 directions ontbreek (u of v) maken we de ander dmv crossproduct met de normal
			if (uDirection.isNewVec())
			{
				uDirection = normal.cross(vDirection);
			}
			if (vDirection.isNewVec())
			{
				vDirection = normal.cross(uDirection);
			}
		}
	}
	
	public IntersectionInfo intersect( Ray r ) 
	{
		float teller = -offset - normal.dot(r.origin);
		float noemer = normal.dot(r.direction);
		//als de noemer 0 is, loopt de ray evenwijdig aan de plane: ofwel er langs of er doorheen.
		//in beide gevallen zie je de plane niet, want de plane is oneindig dun
		if(noemer == 0)
			return new IntersectionInfo(false);
		else
		{
			float t = teller / noemer;
			if(t<0)
				return new IntersectionInfo(false);
			else
			{
				Vec3 location = r.origin.add( r.direction.times(t) );
				//even kijken aan welke kant van de plane we zitten
				//tevens moeten we de direction vector omdraaien, omdat we anders de verkeerde
				//hoek uitrekenen
				float coshoek =  (normal.dot( r.direction.times(-1) ))/ (normal.length() * r.direction.length()) ;
				Vec3 normaltemp;
				if( coshoek < 0)
					normaltemp = normal.times(-1);
				else
					normaltemp = normal;
				
				normaltemp.normalize();
				
				float distance = (location.minus(r.origin)).length();
				
				//berekenen van u/v coordinaten.
				float u = location.dot( uDirection );
				float v = location.dot( vDirection );
				
				return new IntersectionInfo(location, normaltemp, distance, this, u, v);
			}
		}
	}
	
	public boolean hit( Ray r ) 
	{
		float teller = -offset - normal.dot(r.origin);
		float noemer = normal.dot(r.direction);
		//als de noemer 0 is, loopt de ray evenwijdig aan de plane: ofwel er langs of er doorheen.
		//in beide gevallen zie je de plane niet, want de plane is oneindig dun
		if(noemer == 0)
			return false;
		else
		{
			float t = teller / noemer;
			return (t <= 1 && t >= 0);
		}
	}
}