package tracer;

import java.io.IOException;

/**
 * A 3D roundish traceable thingy with a center and a radius
 */
public class Sphere extends Traceable {

	Vec3 origin;
	float radius;
	
	public Sphere( Vec3 o, float r ) {
		origin = new Vec3(o);
		radius = r;	
		material = new Material();
	}
	public Sphere() {
		origin = new Vec3();
		material = new Material();	
	}
	
	public void parse( Parser p ) throws IOException {
		p.parseKeyword( "{" );
		while( !p.tryKeyword("}") && !p.endOfFile() ) {
			
			if( p.tryKeyword("origin") ) {
				origin.parse( p );
			} else if( p.tryKeyword("radius") ) {
				radius = p.parseFloat();
			} else if( p.tryKeyword("material") ) {
				material.parse( p );
			} else {
				System.out.println( p.tokenWasUnexpected() );	
			}
			
		}
	}
	
	public IntersectionInfo intersect( Ray r ) 
	{
		Vec3 d = r.direction;
		Vec3 eMinC = r.origin.minus(origin);
		float a = d.dot(d);
		float b = ( d.times(2) ).dot( eMinC );
		float c = ( eMinC.dot(eMinC) ) - (radius * radius);
		float discriminant = (b * b) - (4 * a * c);
		
		if(discriminant < 0) //er zijn geen oplossingen, ofwel geen raakpunten
		{
			return new IntersectionInfo(false);
		}
		else
		{
			float t1 = (-b - (float)Math.sqrt(discriminant)) / (2*a);
			float t;
			
			if (discriminant == 0) // er is maar 1 raakpunt
			{
				t = t1;
			}
			else // er zijn 2 raakpunten
			{
				float t2 = (-b + (float)Math.sqrt(discriminant)) / (2*a);
				
				if (t1 < 0 && t2 < 0) // de sphere staat achter de camera
					return new IntersectionInfo(false);
				else
				{
					if (t1 < 0 || t2 < 0) // de camera staat in de sphere
						t = Math.max(t1, t2);
					else // de normale situatie, de sphere staat voor de camera
						t = Math.min(t1, t2);
				}
			}
			
			Vec3 location = r.origin.add( r.direction.times(t));
			Vec3 normal = location.minus(origin);
			normal.normalize();
			float distance = t * d.length();
			
			//u,v coordinaten berekenen. dit doen we mbv cylindrical mapping
			double phi   = Math.atan2(origin.z - location.z, location.x - origin.x);
			
			phi -= 0.5f * (float) Math.PI;
			if(phi < 0)
				phi += (float) 2*Math.PI;

			float u = (float) ((phi)/(2*Math.PI));
			float v =  (location.y - origin.y + radius)/(2*radius);
			
			return new IntersectionInfo(location, normal, distance, this, u, v);
		}
	}
	
	public boolean hit( Ray r ) 
	{
		Vec3 d = r.direction;
		Vec3 eMinC = r.origin.minus(origin);
		float a = d.dot(d);
		float b = ( d.times(2) ).dot( eMinC );
		float c = ( eMinC.dot(eMinC) ) - (radius * radius);
		float discriminant = (b * b) - (4 * a * c);
		
		if(discriminant < 0) //er zijn geen oplossingen, ofwel geen raakpunten
		{
			return false;
		}
		else
		{
			float t1 = (-b - (float)Math.sqrt(discriminant)) / (2*a);
			float t2 = (-b + (float)Math.sqrt(discriminant)) / (2*a);
			
			return (t1 <= 1 && t1 >= 0) || (t2 <= 1 && t2 >= 0);
		}
	}
	
}