package tracer;

import java.io.IOException;

public class Triangle extends Traceable
{
	Vec3 a, b, c;
	float u1,v1,u2,v2,u3,v3;
	
	public Triangle(Vec3 p1, Vec3 p2, Vec3 p3)
	{
		this.a = p1;
		this.b = p2;
		this.c = p3;
		material = new Material();
	}
	public Triangle()
	{
		this.a = new Vec3();
		this.b = new Vec3();
		this.c = new Vec3();
		material = new Material();
	}
	
	public void parse( Parser p ) throws IOException {
		p.parseKeyword( "{" );
		while( !p.tryKeyword("}") && !p.endOfFile() ) {
			
			if( p.tryKeyword("p1") ) {
				a.parse( p );
			} else if( p.tryKeyword("u1") ) {
				u1 = p.parseFloat();
			} else if( p.tryKeyword("v1") ) {
				v1 = p.parseFloat();
			} else if( p.tryKeyword("p2") ) {
				b.parse( p );
			} else if( p.tryKeyword("u2") ) {
				u2 = p.parseFloat();
			} else if( p.tryKeyword("v2") ) {
				v2 = p.parseFloat();
			} else if( p.tryKeyword("p3") ) {
				c.parse( p );
			} else if( p.tryKeyword("u3") ) {
				u3 = p.parseFloat();
			} else if( p.tryKeyword("v3") ) {
				v3 = p.parseFloat();
			} else if( p.tryKeyword("material") ) {
				material.parse( p );
			} else {
				System.out.println( p.tokenWasUnexpected() );	
			}
			
		}
	}
	
	public IntersectionInfo intersect( Ray r ) 
	{
		//voor het bepalen van een intersectionpunt met ray-triangle gebruiken we kramers rule. 
		//Dit om ingewikkelde matrix berekeningen te voorkomen. zie ook boek editie 2, blz 207/208
		
		float eiMINhf = ( (c.y - a.y) * -r.direction.z ) - ( -r.direction.y* (c.z - a.z) );
		float gfMINdi = ( -r.direction.x * (c.z - a.z) ) - ( (c.x - a.x) * -r.direction.z );
		float dhMINeg = ( (c.x - a.x) * -r.direction.y ) - ( (c.y - a.y) * -r.direction.x );
		float akMINjb = ( (b.x - a.x) * (r.origin.y - a.y) ) - ( (r.origin.x - a.x) * (b.y - a.y) );
		float jcMINal = ( (r.origin.x - a.x) * (b.z - a.z) ) - ( (b.x - a.x) * (r.origin.z - a.z) );
		float blMINkc = ( (b.y - a.y) * (r.origin.z - a.z) ) - ( (r.origin.y - a.y) * (b.z - a.z) );
		float m = ( (b.x - a.x) * eiMINhf ) + ( (b.y - a.y) * gfMINdi ) + ( (b.z - a.z) * dhMINeg );
		
		float t = -1 * ( ( ( (c.z - a.z) * akMINjb ) + ( (c.y - a.y) * jcMINal ) + ( (c.x - a.x) * blMINkc ) ) / m );
		if ( t < 0 ) // intersection is achter de camera (we kijken niet of de intersection te ver weg is)
			return new IntersectionInfo (false);
		
		float gamma = ( ( ( (-r.direction.z) * akMINjb ) + ( (-r.direction.y) * jcMINal ) + ( (-r.direction.x) * blMINkc ) ) / m );
		if (gamma < 0 || gamma > 1)
			return new IntersectionInfo (false); // intersection wel op de plane maar niet in de triangle
		
		float beta = ( ( (r.origin.x - a.x) * eiMINhf ) + ( (r.origin.y - a.y) * gfMINdi ) + ( (r.origin.z - a.z) * dhMINeg ) ) / m ;
	
		if (beta < 0 || beta > (1 - gamma))
			return new IntersectionInfo (false); // intersection wel op de plane maar niet in de triangle
	
		//nu hebben we een geldige intersection
		Vec3 location = r.origin.add( r.direction.times(t) );
		
		/*
		 *even kijken aan welke kant van de triangle we zitten
		 *tevens moeten we de direction vector omdraaien, omdat we anders de verkeerde
		 *hoek uitrekenen. maar eerst maar eens de normal berekenen:
		 */
		Vec3 normal = (a.minus(b)).cross( c.minus(b));
		float coshoek =  (normal.dot( r.direction.times(-1) ))/ (normal.length() * r.direction.length()) ;
		Vec3 normaltemp;
		if( coshoek < 0)
			normaltemp = normal.times(-1);
		else
			normaltemp = normal;
		normaltemp.normalize();
		
		//uv coordinates voor texture mapping berekenen
		float u = beta*(u2-u1) + gamma*(u3-u1) + u1;
		float v = beta*(v2-v1) + gamma*(v3-v1) + v1;
		
		float distance = (location.minus(r.origin)).length();
		return new IntersectionInfo(location, normaltemp, distance, this, u, v);
	}
	
	public boolean hit( Ray r )
	{
		//zie ook intersect
		float eiMINhf = ( (c.y - a.y) * -r.direction.z ) - ( -r.direction.y* (c.z - a.z) );
		float gfMINdi = ( -r.direction.x * (c.z - a.z) ) - ( (c.x - a.x) * -r.direction.z );
		float dhMINeg = ( (c.x - a.x) * -r.direction.y ) - ( (c.y - a.y) * -r.direction.x );
		float akMINjb = ( (b.x - a.x) * (r.origin.y - a.y) ) - ( (r.origin.x - a.x) * (b.y - a.y) );
		float jcMINal = ( (r.origin.x - a.x) * (b.z - a.z) ) - ( (b.x - a.x) * (r.origin.z - a.z) );
		float blMINkc = ( (b.y - a.y) * (r.origin.z - a.z) ) - ( (r.origin.y - a.y) * (b.z - a.z) );
		float m = ( (b.x - a.x) * eiMINhf ) + ( (b.y - a.y) * gfMINdi ) + ( (b.z - a.z) * dhMINeg );
		
		float t = -1 * ( ( ( (c.z - a.z) * akMINjb ) + ( (c.y - a.y) * jcMINal ) + ( (c.x - a.x) * blMINkc ) ) / m );
		if ( t < 0 || t > 1) // intersection is achter de camera (we kijken niet of de intersection te ver weg is)
			return false;
		
		float gamma = ( ( ( (-r.direction.z) * akMINjb ) + ( (-r.direction.y) * jcMINal ) + ( (-r.direction.x) * blMINkc ) ) / m );
		if (gamma < 0 || gamma > 1)
			return false;
		
		float beta = ( ( (r.origin.x - a.x) * eiMINhf ) + ( (r.origin.y - a.y) * gfMINdi ) + ( (r.origin.z - a.z) * dhMINeg ) ) / m ;

		if (beta < 0 || beta > (1 - gamma))
			return false;
		
		//als we tot hier komen, hebben we dus een geldige hit
		return true;
	}
}
